from derivation_method.action import GetData
from derivation_method.derivation_method import OnlineDerivationMethod, derivation_method_factory, DictDerivationMethod, RdfDerivationMethod, SnippetDerivationMethod, DerivationMethod
